from itertools import combinations
from collections import defaultdict
import scipy
import scipy.spatial
import scipy.spatial.distance
import scipy.special
import multiprocessing

def read_ply(file):
    """
    Support function for reading in points from a PLY/ASCII 3d mesh file.
    """
    dragonPts = None
    dragonF=open(file, 'r')
    i = 0
    maxi = None
    for l in dragonF:
        line = l.split(' ')
        if line[0] == 'element' and line[1] == 'vertex':
            maxi = int(line[2])
            dragonPts = scipy.zeros((maxi,3))
        elif l[0] in ['0','1','2','3','4','5','6','7','8','9'] and line[0] != '3':
            dragonPts[i,:] = [float(line[j]) for j in range(3)]
            i += 1
        elif i == maxi:
            break
        else:
            continue
    dragonF.close()
    return dragonPts


def nbhd_graph_edges(dpts, eps, mineps=None):
    """
    Graph construction helper function. Given points in an NxD matrix (dpts),
    and a maximal distance eps, the function nbhd_graph_edges returns a tuple
    with contents:
    a dictionary filtervalue -> edge-list;
    a dictionary edge -> filtervalue;
    a dictionary v -> lower (indexed) neighborhood of v

    If mineps is given, the filtration values will be quantized to even
    multiples of mineps.
    """
    edges = defaultdict(list)
    nbhd = defaultdict(set)
    fedge = {}
    dists = scipy.spatial.distance.squareform(scipy.spatial.distance.pdist(dpts))

    pairs = (dists<eps).nonzero()
    for p in range(len(pairs[0])):
        i,j = pairs[0][p], pairs[1][p]
        d = dists[i,j]
        if i >= j: continue
        if mineps: d=mineps*int(d/mineps)
        edges[d].append((i,j))
        fedge[(i,j)] = d
        nbhd[j].add(i)
    return (edges,fedge,nbhd)

def f(splx, fedge):
    """
    Given a simplex splx and a dictionary edge -> filtervalue as returned
    by nbhd_graph_edges, returns the maximal filtervalue of edges in the
    simplex.
    """
    if len(splx) == 1: return 0
    if len(splx) == 2:
        try:
            return fedge[tuple(sorted(splx))]
        except:
            return float('inf')
    fvals = [fedge[(i,j)] for (i,j) in combinations(sorted(splx), 2)]
    try: return max(fvals)
    except: return float('inf')

def add_cofaces(GV, Gnb, k, tau, N):
    """
    The function add_cofaces as defined by Zomorodian (2010).
    """
    ret = [tau]
    if len(tau) <= k:
        for n in N:
            sigma = tau.union({n})
            M = Gnb[n].intersection(N)
            ret.extend(add_cofaces(GV, Gnb, k, sigma, M))
    return ret

def incremental(GV, Gnb, k, fedge):
    """
    Classical incremental algorithm for Vietoris-Rips complexes, finishing
    with a sorting of the simplices in filtration value order.
    Specified in Zomorodian (2010).
    """
    V = []
    for u in GV:
        V.extend(add_cofaces(GV, Gnb, k, {u}, Gnb[u]))
    return sorted(V, key=lambda spx: f(spx, fedge))

def incremental_order(GV, Gnb, k, fedge, edges):
    """
    In-order version of the incremental algorithm for Vietoris-Rips
    complexes.
    """
    V = [set([v]) for v in GV]
    iGV = []
    iGnb = defaultdict(set)
    for fv in sorted(edges.keys()):
        for e in edges[fv]:
            iGV.append(e)
            iGnb[max(e)].add(min(e))
            iGnb[min(e)].add(max(e))
            V.extend(add_cofaces(iGV, iGnb, k, set(e), iGnb[min(e)].intersection(iGnb[max(e)])))
    return V

def edge_partition(edges, fedge, nbins=20, bdry=None):
    """
    Divides the filtration values for the graph specified through edges and
    fedge (as returned as the first and second tuple member from nbhd_graph_edges)
    into decaying sized bins. nbins is the number of bins requested,
    and the optional bdry parameter lets you control the bin boundaries
    for the partition.

    Returns a list of lists of filtration values.
    """
    if not bdry:
        step = len(fedge)/nbins + 1
        bdry = [x for x in scipy.unique(list(scipy.cumsum([len(fedge)*.25*.75**(k) for k in range(nbins)]))+[len(fedge)])
            if 0 < x <= len(fedge)]
    bins = []
    tmpbin = []
    sum = 0
    i = 0
    for f in sorted(edges.keys()):
        tmpbin.append(f)
        sum += len(edges[f])
        if sum > bdry[i]:
            bins.append(tmpbin)
            tmpbin=[]
            i += 1
    bins.append(tmpbin)
    return bins

def incremental_helper(GV, Gnb, k, fedge, edges, bin):
    """
    Helper function for distributing the Vietoris-Rips construction.
    Given a list of filtration values to handle, this function bootstraps
    the preceding graph construction, and then runs the in-order Vietoris-Rips
    construction and returns the additional simplices induced by this set of
    edges.
    """
    V = []
    minfv = bin[0]
    iGV = []
    iGnb = defaultdict(set)
    for fv in edges:
        if fv < minfv:
            for e in edges[fv]:
                iGV.append(e)
                iGnb[max(e)].add(min(e))
                iGnb[min(e)].add(max(e))

    for fv in bin:
        for e in edges[fv]:
            iGV.append(e)
            iGnb[max(e)].add(min(e))
            iGnb[min(e)].add(max(e))
            V.extend(add_cofaces(iGV, iGnb, k, set(e), iGnb[min(e)].intersection(iGnb[max(e)])))

    return V


def incremental_order_partition(GV, Gnb, k, fedge, edges):
    """
    Example implementation of distributed computation of a Vietoris-Rips complex
    using the in-order incremental construction algorithm. This particular
    implementation uses the Python multiprocessing library for distributing
    the computation on several processes.
    """
    V = [set([v]) for v in GV]
    binres=[]
    bins = edge_partition(edges,fedge)
    for i in range(len(bins)):
        binres.append(multiprocessing.Process(target=incremental_helper, args=(GV, Gnb, k, fedge, edges, bins[i])))
        binres[-1].start()
    for j in binres:
        j.join()
    ret = V + [k for l in [br.get() for br in binres] for k in l]
    return ret

def incremental_order_partition_serial(GV, Gnb, k, fedge, edges):
    """
    Serial implementation of incremental_order_partition to verify correctness
    of the implementation.
    """
    V = [set([v]) for v in GV]
    bins = edge_partition(edges, fedge)
    binres=[V]
    for i in range(len(bins)):
        binres.append(incremental_helper(GV, Gnb, k, fedge, edges, bins[i]))
    return [s for lst in binres for s in lst]
